/*
    Project: TCGINA
    Version: V1.18
    File   : tcmisc.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcpriv.h"
#include <sddl.h>
#include <tchar.h>

/*-----------------------------------------------------------------------------
// String functions                                                          */

BOOL WINAPI TC_AppendString(LPTSTR pszDest, DWORD nSize, LPCTSTR pszSrc)
{
    size_t nDestLen, nSrcLen;

    TC_Assert(pszDest);
    TC_Assert(pszSrc);

    if (pszDest == NULL || pszSrc == NULL) return FALSE;

    nSrcLen = _tcslen(pszSrc);
    nDestLen = _tcslen(pszDest);
    if (nSrcLen + nDestLen >= nSize) return FALSE;

    memcpy(pszDest + nDestLen, pszSrc, (nSrcLen+1)*sizeof(TCHAR));

    return TRUE;
}

BOOL WINAPI TC_CopyString(LPTSTR pszDest, DWORD nSize, LPCTSTR pszSrc)
{
    size_t nSrcLen;

    TC_Assert(pszDest);
    TC_Assert(pszSrc);

    if (pszDest == NULL || pszSrc == NULL) return FALSE;

    nSrcLen = _tcslen(pszSrc);
    if (nSrcLen >= nSize) return FALSE;

    memcpy(pszDest, pszSrc, (nSrcLen+1)*sizeof(TCHAR));

    return TRUE;
}

void WINAPI TC_TrimString(LPTSTR pszString)
{
    LPTSTR pszStart;
    LPTSTR pszEnd;

    TC_Assert(pszString);

    if (pszString == NULL) return;

    for (pszStart = pszString; _istspace(*pszStart); pszStart++) ;
    pszEnd = pszStart + _tcslen(pszStart);

    while (pszStart != pszEnd) {
        pszEnd--;
        if (!_istspace(*pszEnd)) {
            size_t n = (pszEnd - pszStart) + 1;
            if (pszStart != pszString && n) {
                memmove(pszString, pszStart, n * sizeof(TCHAR));
            }
            pszString[n] = TEXT('\0');
            return;
        }
    }

    *pszString = TEXT('\0');
}

BOOL WINAPI TC_ReplaceString(LPTSTR pszDest, DWORD nDestSize,
  LPCTSTR pszSearch, LPCTSTR pszReplace)
{
    size_t nSearchSize;
    size_t nReplaceSize;
    size_t nSize;

    TC_Assert(pszDest);
    TC_Assert(pszSearch);
    TC_Assert(pszReplace);

    if (pszDest == NULL || pszSearch == NULL || pszReplace == NULL) {
        return FALSE;
    }

    nSize = _tcslen(pszDest);
    nSearchSize = _tcslen(pszSearch);
    nReplaceSize = _tcslen(pszReplace);

    for (;;) {

        LPTSTR pszTemp = _tcsstr(pszDest, pszSearch);
        if (pszTemp == NULL) return TRUE;

        nSize += nReplaceSize;
        nSize -= nSearchSize;
        if (nSize >= nDestSize) return FALSE;

        if (nSearchSize > nReplaceSize) {
            size_t n = nSearchSize - nReplaceSize;
            memmove(pszTemp, pszTemp + n,
              (_tcslen(pszTemp + n) + 1)*sizeof(TCHAR));
        } else if (nSearchSize < nReplaceSize) {
            memmove(pszTemp + nReplaceSize, pszTemp + nSearchSize,
              (_tcslen(pszTemp + nSearchSize) + 1)*sizeof(TCHAR));
        }
        memcpy(pszTemp, pszReplace, nReplaceSize*sizeof(TCHAR));
        pszDest = pszTemp + nReplaceSize;
    }
}

LPCTSTR WINAPI TC_GetNextPathName(LPCTSTR pszSearchPath, LPTSTR pszDest,
  DWORD nDestSize)
{
    TC_Assert(pszDest);
    TC_Assert(nDestSize);

    if (pszSearchPath == NULL) {
        *pszDest = TEXT('\0');
        return NULL;
    }

    for(;;) {

        LPCTSTR pszTemp;
        size_t  nSize;

        while (_istspace(*pszSearchPath) || *pszSearchPath == TEXT(';')) {
            pszSearchPath++;
        }

        if (*pszSearchPath == '\"') {
            pszSearchPath++;
            while (_istspace(*pszSearchPath)) pszSearchPath++;
            pszTemp = _tcschr(pszSearchPath, TEXT('\"'));
        } else {
            pszTemp = _tcschr(pszSearchPath, TEXT(';'));
        }

        if (pszTemp == NULL) {
            if (!TC_CopyString(pszDest, nDestSize, pszSearchPath)) {
                *pszDest = TEXT('\0');
            }
            TC_TrimString(pszDest);
            TC_RemoveTerminatingBackslash(pszDest);
            return NULL;
        }

        nSize = pszTemp - pszSearchPath;
        if (nSize > 0 && nSize <= nDestSize) {
            memcpy(pszDest, pszSearchPath, nSize*sizeof(TCHAR));
            pszDest[nSize] = TEXT('\0');
            TC_TrimString(pszDest);
            TC_RemoveTerminatingBackslash(pszDest);
            if (*pszDest) return pszTemp + 1;
        }

        pszSearchPath = pszTemp + 1;
    }
}

/*-----------------------------------------------------------------------------
// Registry functions                                                        */

/* Return value must be freed with TC_Free() */
LPCTSTR WINAPI TC_LoadRegistryString(HKEY hKey, LPCTSTR pszValueName)
{
    DWORD  dwType;
    DWORD  cbSize;
    LPTSTR pszDest;

    TC_Assert(pszValueName);

    if (pszValueName == NULL) return NULL;

    cbSize = 0;
    RegQueryValueEx(hKey, pszValueName, NULL, &dwType, NULL, &cbSize);
    if (cbSize == 0) return NULL;

    pszDest = TC_Alloc(cbSize);
    if (pszDest == NULL) return NULL;
    if (RegQueryValueEx(hKey, pszValueName, NULL, &dwType, (LPBYTE) pszDest,
      &cbSize) != ERROR_SUCCESS) {
        TC_Free(pszDest);
        return NULL;
    }

    switch(dwType) {

    case REG_SZ:
         return pszDest;

    case REG_EXPAND_SZ: {

        LPTSTR pszExpandedDest;
        DWORD  nExpandedSize;

        if (*pszDest == TEXT('\0')) return pszDest;

        nExpandedSize = ExpandEnvironmentStrings(pszDest, NULL, 0);
        if (nExpandedSize == 0) return pszDest;

        pszExpandedDest = (LPTSTR) TC_Alloc(nExpandedSize*sizeof(TCHAR));
        if (pszExpandedDest == NULL) return pszDest;

        if (ExpandEnvironmentStrings(pszDest, pszExpandedDest,
          nExpandedSize) == 0) {
            TC_Free(pszExpandedDest);
            return pszDest;
        }

        TC_Free(pszDest);
        return pszExpandedDest; }
    }

    TC_Free(pszDest);

    return NULL;
}

BOOL WINAPI TC_ReadRegistryString(HKEY hKey, LPCTSTR pszValueName,
  LPTSTR pszDest, DWORD nSize)
{
    LPCTSTR pszValue;

    TC_Assert(pszValueName);
    TC_Assert(pszDest);

    pszValue = TC_LoadRegistryString(hKey, pszValueName);
    if (pszValue == NULL) return FALSE;

    if (_tcslen(pszValue) >= nSize) {
        TC_Free(pszValue);
        return FALSE;
    }

    _tcscpy(pszDest, pszValue);
    TC_Free(pszValue);

    return TRUE;
}

BOOL WINAPI TC_GetRegistryString(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName, LPTSTR pszDest, DWORD nSize)
{
    HKEY hKey;

    TC_Assert(pszKeyName);
    TC_Assert(pszValueName);
    TC_Assert(pszDest);

    if (RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey) ==
      ERROR_SUCCESS) {
        if (TC_ReadRegistryString(hKey, pszValueName, pszDest, nSize)) {
            RegCloseKey(hKey);
            return TRUE;
        }
        RegCloseKey(hKey);
    }

    return FALSE;
}

/* Return value must be freed with TC_Free() */
LPCTSTR WINAPI TC_LoadRegistryMultiString(HKEY hKey, LPCTSTR pszValueName,
  PDWORD pdwType)
{
    DWORD  dwType;
    DWORD  cbSize;
    DWORD  cbExtSize;
    LPTSTR pszDest;

    TC_Assert(pszValueName);

    if (pszValueName == NULL) return NULL;

    cbSize = 0;
    RegQueryValueEx(hKey, pszValueName, NULL, &dwType, NULL, &cbSize);
    if (cbSize == 0) return NULL;

    cbExtSize = cbSize + sizeof(TCHAR);
    pszDest = TC_Alloc(cbExtSize);
    if (pszDest == NULL) return NULL;
    memset(pszDest, '\0', cbExtSize);
    if (RegQueryValueEx(hKey, pszValueName, NULL, &dwType, (LPBYTE) pszDest,
      &cbSize) != ERROR_SUCCESS) {
        TC_Free(pszDest);
        return NULL;
    }

    if (pdwType) {
        *pdwType = dwType;
    }

    switch(dwType) {

    case REG_SZ:
    case REG_MULTI_SZ:
         return pszDest;

    case REG_EXPAND_SZ: {

        LPTSTR pszExpandedDest;
        DWORD  nExpandedSize;

        if (*pszDest == TEXT('\0')) return pszDest;

        nExpandedSize = ExpandEnvironmentStrings(pszDest, NULL, 0);
        if (nExpandedSize == 0) return pszDest;

        pszExpandedDest = (LPTSTR) TC_Alloc(nExpandedSize*sizeof(TCHAR));
        if (pszExpandedDest == NULL) return pszDest;

        if (ExpandEnvironmentStrings(pszDest, pszExpandedDest,
          nExpandedSize) == 0) {
            TC_Free(pszExpandedDest);
            return pszDest;
        }

        TC_Free(pszDest);
        return pszExpandedDest; }
    }

    TC_Free(pszDest);

    return NULL;
}

static DWORD TC_GetMultiStringSize(LPCTSTR pszMultiString)
{
    LPCTSTR pszTemp = pszMultiString;

    TC_Assert(pszMultiString);

    if (*pszTemp == TEXT('\0')) {
        return 2;
    }

    while (*pszTemp) {
        pszTemp += _tcslen(pszTemp) + 1;
    }

    return ((DWORD)(pszTemp - pszMultiString)) + 1;
}

BOOL WINAPI TC_ReadRegistryMultiString(HKEY hKey, LPCTSTR pszValueName,
  LPTSTR pszDest, DWORD nSize, PDWORD pdwType)
{
    LPCTSTR pszValue;
    DWORD   nReqSize;

    TC_Assert(pszValueName);
    TC_Assert(pszDest);

    pszValue = TC_LoadRegistryMultiString(hKey, pszValueName, pdwType);
    if (pszValue == NULL) return FALSE;

    nReqSize = TC_GetMultiStringSize(pszValue);
    if (nReqSize >= nSize) {
        TC_Free(pszValue);
        return FALSE;
    }

    memcpy(pszDest, pszValue, nReqSize * sizeof(TCHAR));
    TC_Free(pszValue);

    return TRUE;
}

BOOL WINAPI TC_GetRegistryMultiString(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName, LPTSTR pszDest, DWORD nSize, PDWORD pdwType)
{
    HKEY hKey;

    TC_Assert(pszKeyName);
    TC_Assert(pszValueName);
    TC_Assert(pszDest);

    if (RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey) ==
      ERROR_SUCCESS) {
        if (TC_ReadRegistryMultiString(hKey, pszValueName, pszDest, nSize,
          pdwType)) {
            RegCloseKey(hKey);
            return TRUE;
        }
        RegCloseKey(hKey);
    }

    return FALSE;
}

BOOL WINAPI TC_ReadRegistryStringList(HKEY hKey, LPCTSTR pszValueName,
  LPTSTR pszDest, DWORD nSize)
{
    DWORD   dwType;
    LPCTSTR pszPath;
    LPTSTR  pszTemp;

    if (!TC_ReadRegistryMultiString(hKey, pszValueName, pszDest, nSize,
      &dwType)) {
        return FALSE;
    }

    if (dwType == REG_MULTI_SZ) {
        return TRUE;
    }

    pszPath = pszTemp = pszDest;
    for (;;) {
        TCHAR  szPath[MAX_PATH];
        size_t nLen;

        pszPath = TC_GetNextPathName(pszPath, szPath, MAX_PATH);
        if (*szPath == TEXT('\0')) {
            *pszTemp = TEXT('\0');
            break;
        }
        nLen = _tcslen(szPath) + 1;
        memcpy(pszTemp, szPath, nLen*sizeof(TCHAR));
        pszTemp += nLen;
    }

    return TRUE;
}

BOOL WINAPI TC_GetRegistryStringList(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName, LPTSTR pszDest, DWORD nSize)
{
    HKEY hKey;

    TC_Assert(pszKeyName);
    TC_Assert(pszValueName);
    TC_Assert(pszDest);

    if (RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey) ==
      ERROR_SUCCESS) {
        if (TC_ReadRegistryStringList(hKey, pszValueName, pszDest, nSize)) {
            RegCloseKey(hKey);
            return TRUE;
        }
        RegCloseKey(hKey);
    }

    return FALSE;
}

BOOL WINAPI TC_ReadRegistryInteger(HKEY hKey, LPCTSTR pszValueName,
  DWORD *pdwValueData)
{
    DWORD dwType;
    DWORD dwData;
    DWORD cbSize;

    TC_Assert(pszValueName);
    TC_Assert(pdwValueData);

    if (pdwValueData == NULL) return FALSE;
    dwData = 0;
    cbSize = sizeof(dwData);

    if (RegQueryValueEx(hKey, pszValueName, NULL, &dwType,
      (LPBYTE)&dwData, &cbSize) != ERROR_SUCCESS) return FALSE;

    switch(dwType) {
    case REG_DWORD:
        *pdwValueData = dwData;
        return TRUE;
    }

    return FALSE;
}

DWORD WINAPI TC_GetRegistryInteger(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName, DWORD dwValueData)
{
    HKEY hKey;

    TC_Assert(pszKeyName);
    TC_Assert(pszValueName);

    if (RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey) ==
      ERROR_SUCCESS) {
        TC_ReadRegistryInteger(hKey, pszValueName, &dwValueData);
        RegCloseKey(hKey);
    }

    return dwValueData;
}

/*-----------------------------------------------------------------------------
// System functions                                                          */

/* Return value must be freed with TC_Free() */
LPCTSTR WINAPI TC_GetStringSid(LPCTSTR pszUserName)
{
    DWORD cbSid = 0;
    DWORD nReferencedDomainNameSize = 0;
    SID_NAME_USE SidNameUse;

    PBYTE pSid;
    LPTSTR pszReferencedDomainName;
    LPTSTR pszSidName;

    TC_Assert(pszUserName);

    if (pszUserName == NULL) return NULL;

    LookupAccountName(NULL, pszUserName, NULL, &cbSid, NULL,
      &nReferencedDomainNameSize, &SidNameUse);

    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) return NULL;
    if (cbSid == 0 || nReferencedDomainNameSize == 0) return NULL;

    pSid = (PBYTE) TC_Alloc(cbSid);
    if (pSid == NULL) return NULL;

    pszReferencedDomainName = (LPTSTR) TC_Alloc(nReferencedDomainNameSize *
      sizeof(TCHAR));
    if (pszReferencedDomainName == NULL) {
        TC_Free(pSid);
        return NULL;
    }

    if (!LookupAccountName(NULL, pszUserName, pSid, &cbSid,
      pszReferencedDomainName, &nReferencedDomainNameSize, &SidNameUse)) {
        TC_Free(pSid);
        TC_Free(pszReferencedDomainName);
        return NULL;
    }
    TC_Free(pszReferencedDomainName);

    if (!ConvertSidToStringSid(pSid, &pszSidName)) {
        TC_Free(pSid);
        return NULL;
    }
    TC_Free(pSid);

    return (LPCTSTR)TC_ConvertLocalAllocBlock(pszSidName,
      (_tcslen(pszSidName)+1)*sizeof(TCHAR));
}

/* Return value must be freed with TC_Free() */
LPCTSTR WINAPI TC_GetUserName(LPCTSTR pszSid)
{
    PSID pSid;
    LPTSTR pszUserName, pszReferencedDomainName;
    DWORD nUserNameSize = 0;
    DWORD nReferencedDomainNameSize = 0;
    SID_NAME_USE SidNameUse;

    TC_Assert(pszSid);

    if (pszSid == NULL) return NULL;

    if (!ConvertStringSidToSid(pszSid, &pSid)) return NULL;
    LookupAccountSid(NULL, pSid, NULL, &nUserNameSize, NULL,
      &nReferencedDomainNameSize, &SidNameUse);
    pszUserName = (LPTSTR) TC_Alloc(nUserNameSize * sizeof(TCHAR));
    pszReferencedDomainName = (LPTSTR) TC_Alloc(
      nReferencedDomainNameSize * sizeof(TCHAR));
    if (pszUserName == NULL || pszReferencedDomainName == NULL) {
        LocalFree((PVOID)pSid);
        TC_Free(pszReferencedDomainName);
        TC_Free(pszUserName);
        return NULL;
    }

    if (!LookupAccountSid(NULL, pSid, pszUserName, &nUserNameSize,
      pszReferencedDomainName, &nReferencedDomainNameSize, &SidNameUse)) {
        LocalFree((PVOID)pSid);
        TC_Free(pszReferencedDomainName);
        TC_Free(pszUserName);
        return NULL;
    }

    LocalFree((PVOID)pSid);
    TC_Free(pszReferencedDomainName);

    return pszUserName;
}

/* Return value must be freed with TC_Free() */
static LPCTSTR WINAPI TC_FindProfileStringSid(LPCTSTR pszUserName)
{
    HKEY hKey;

    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, PROFILE_LIST_KEY_NAME, 0, KEY_READ,
      &hKey) == ERROR_SUCCESS) {
        TCHAR szKeyName[MAX_PATH];
        DWORD dwIndex;
        DWORD dwSize;

        dwIndex = 0;
        dwSize  = MAX_PATH;
        while (RegEnumKeyEx(hKey, dwIndex, szKeyName, &dwSize,
          NULL, NULL, NULL, NULL) == ERROR_SUCCESS) {
            LPCTSTR pszRegistryUserName;
            pszRegistryUserName = TC_GetUserName(szKeyName);
            if (pszRegistryUserName) {
                if (_tcsicmp(pszUserName, pszRegistryUserName) == 0) {
                    size_t cbResultSize;
                    LPTSTR pszResult;

                    TC_Free(pszRegistryUserName);
                    RegCloseKey(hKey);

                    cbResultSize = (_tcslen(szKeyName) + 1) * sizeof(TCHAR);
                    pszResult = (LPTSTR) TC_Alloc(cbResultSize);
                    if (pszResult == NULL) {
                        return NULL;
                    }
                    memcpy(pszResult, szKeyName, cbResultSize);
                    return pszResult;
                }
                TC_Free(pszRegistryUserName);
            }
            dwSize = MAX_PATH;
            dwIndex++;
        }
        RegCloseKey(hKey);
    }

    return NULL;
}

static BOOL TC_FindProfileImagePath(LPCTSTR pszSid, LPTSTR pszProfileImagePath,
  DWORD nSize)
{
    LPTSTR  pszRegKeyName;
    size_t  nSidSize;

    TC_Assert(pszSid);
    TC_Assert(pszProfileImagePath);

    nSidSize = _tcslen(pszSid);

    pszRegKeyName = (LPTSTR) TC_Alloc(
      (nSidSize + (PROFILE_LIST_KEY_NAME_BS_LEN+1)) * sizeof(TCHAR));
    if (pszRegKeyName == NULL) {
        TC_Free(pszSid);
        return FALSE;
    }
    _tcscpy(pszRegKeyName, PROFILE_LIST_KEY_NAME_BS);
    _tcscpy(pszRegKeyName + PROFILE_LIST_KEY_NAME_BS_LEN, pszSid);

    if (TC_GetRegistryString(HKEY_LOCAL_MACHINE, pszRegKeyName,
      TEXT("ProfileImagePath"), pszProfileImagePath, nSize)) {
        TC_RemoveTerminatingBackslash(pszProfileImagePath);
        TC_Free((PVOID)pszRegKeyName);
        return TRUE;
    }
    TC_Free((PVOID)pszRegKeyName);

    return FALSE;
}

/* Return value must be freed with TC_Free() */
LPCTSTR WINAPI TC_GetProfileStringSid(LPCTSTR pszUserName)
{
    TCHAR   szProfileImagePath[MAX_PATH];
    LPCTSTR pszSid;

    pszSid = TC_GetStringSid(pszUserName);
    if (pszSid) {
        if (TC_FindProfileImagePath(pszSid, szProfileImagePath, MAX_PATH)) {
            return pszSid;
        }
        TC_DbgPrintf(TEXT("Cannot find profile for SID=%s\n"), pszSid);
        TC_Free(pszSid);
    }
    pszSid = TC_FindProfileStringSid(pszUserName);
    if (pszSid) {
        TC_DbgPrintf(TEXT("SID found by enumerating ProfileList: %s\n"), pszSid);
        if (TC_FindProfileImagePath(pszSid, szProfileImagePath, MAX_PATH)) {
            return pszSid;
        }
        TC_DbgPrintf(TEXT("Cannot find profile for SID=%s\n"), pszSid);
        TC_Free(pszSid);
    }

    return NULL;
}

BOOL WINAPI TC_GetProgramFilesDirectory(LPTSTR pszProgramFilesDirectory,
  DWORD nSize)
{
    DWORD nRequiredSize;

    TC_Assert(pszProgramFilesDirectory);

    if (TC_GetRegistryString(HKEY_LOCAL_MACHINE,
      TEXT("SOFTWARE\\Microsoft\\Windows\\CurrentVersion"),
      TEXT("ProgramFilesDir"), pszProgramFilesDirectory, nSize)) {
        TC_RemoveTerminatingBackslash(pszProgramFilesDirectory);
        return TRUE;
    }

    if (TC_GetRegistryString(HKEY_LOCAL_MACHINE,
      TEXT("SOFTWARE\\Microsoft\\Windows\\CurrentVersion"),
      TEXT("ProgramFilesPath"), pszProgramFilesDirectory, nSize)) {
        TC_RemoveTerminatingBackslash(pszProgramFilesDirectory);
        return TRUE;
    }

    nRequiredSize = GetEnvironmentVariable(TEXT("ProgramFiles"),
      pszProgramFilesDirectory, nSize);

    if (nRequiredSize >= nSize) return FALSE;

    TC_RemoveTerminatingBackslash(pszProgramFilesDirectory);

    return TRUE;
}

BOOL WINAPI TC_GetProfilesDirectory(LPTSTR pszProfilesDirectory, DWORD nSize)
{
    TC_Assert(pszProfilesDirectory);

    if (TC_GetRegistryString(HKEY_LOCAL_MACHINE,
      PROFILE_LIST_KEY_NAME, TEXT("ProfilesDirectory"),
      pszProfilesDirectory, nSize)) {
        TC_RemoveTerminatingBackslash(pszProfilesDirectory);
        return TRUE;
    }

    return FALSE;
}

#if 0
BOOL WINAPI TC_GetProfileImagePath(LPCTSTR pszUserName,
  LPTSTR pszProfileImagePath, DWORD nSize)
{
    LPTSTR  pszRegKeyName;
    LPCTSTR pszSid;
    size_t  nSidSize;

    TC_Assert(pszUserName);
    TC_Assert(pszProfileImagePath);

    if ((pszSid = TC_GetStringSid(pszUserName)) == NULL) return FALSE;

    nSidSize = _tcslen(pszSid);

    pszRegKeyName = (LPTSTR) TC_Alloc(
      (nSidSize + (PROFILE_LIST_KEY_NAME_BS_LEN+1)) * sizeof(TCHAR));
    if (pszRegKeyName == NULL) {
        TC_Free(pszSid);
        return FALSE;
    }
    _tcscpy(pszRegKeyName, PROFILE_LIST_KEY_NAME_BS);
    _tcscpy(pszRegKeyName + PROFILE_LIST_KEY_NAME_BS_LEN, pszSid);
    TC_Free(pszSid);

    if (TC_GetRegistryString(HKEY_LOCAL_MACHINE, pszRegKeyName,
      TEXT("ProfileImagePath"), pszProfileImagePath, nSize)) {
        TC_RemoveTerminatingBackslash(pszProfileImagePath);
        TC_Free((PVOID)pszRegKeyName);
        return TRUE;
    }

    TC_Free((PVOID)pszRegKeyName);

    return FALSE;
}
#endif

/*-----------------------------------------------------------------------------
// Filename functions                                                        */

void WINAPI TC_RemoveTerminatingBackslash(LPTSTR pszPath)
{
    TC_Assert(pszPath);

    if (pszPath == NULL || *pszPath == TEXT('\0')) return;

    pszPath = pszPath + (_tcslen(pszPath) - 1);

    if (*pszPath == TEXT('\\') || *pszPath == TEXT('/')) *pszPath = TEXT('\0');
}

BOOL WINAPI TC_NormalizeFileName(LPTSTR pszDest, DWORD nDestSize,
  LPCTSTR pszSrc)
{
    size_t nSrcSize;
    LPTSTR p;

    TC_Assert(pszSrc);
    TC_Assert(pszDest);

    if (nDestSize == 0) return FALSE;
    if (pszDest) *pszDest = TEXT('\0');
    if (pszSrc == 0) return FALSE;
    nSrcSize = _tcslen(pszSrc);
    if (nSrcSize >= nDestSize) return FALSE;

    memcpy(pszDest, pszSrc, (nSrcSize+1)*sizeof(TCHAR));

    p = pszDest;
    for (p = pszDest; (p = _tcschr(p, TEXT('/'))) != NULL; p++) {
        *p = TEXT('\\');
    }

    return TRUE;
}

BOOL WINAPI TC_FileExists(LPCTSTR pszFileName)
{
    DWORD dwAttr;

    TC_Assert(pszFileName);

    dwAttr = GetFileAttributes(pszFileName);
    if (dwAttr == 0xffffffff) {
        HANDLE hFile;

        hFile = CreateFile(pszFileName, 0, FILE_SHARE_READ|FILE_SHARE_WRITE,
          NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
            CloseHandle(hFile);
            TC_DbgPrintf(TEXT("Warning: Cannot access directory of ")
              TEXT("\"%s\" (Please verify its DACL!)\n"), pszFileName);
            return TRUE;
        } else if (GetLastError() == ERROR_ACCESS_DENIED) {
            TC_DbgPrintf(TEXT("Warning: Cannot access ")
                TEXT("\"%s\" (Please verify its DACL!)\n"), pszFileName);
            return TRUE;
        } else {
            TC_DbgPrintf(TEXT("Cannot find file \"%s\"!\n"), pszFileName);
        }
        return FALSE;
    }

    return (dwAttr & FILE_ATTRIBUTE_DIRECTORY) == 0;
}

BOOL WINAPI TC_PathExists(LPCTSTR pszFileName)
{
    DWORD dwAttr;

    TC_Assert(pszFileName);

    dwAttr = GetFileAttributes(pszFileName);
    if (dwAttr == 0xffffffff) {
        HANDLE hFile;

        hFile = CreateFile(pszFileName, 0, FILE_SHARE_READ|FILE_SHARE_WRITE,
          NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if (hFile != INVALID_HANDLE_VALUE) {
            CloseHandle(hFile);
            TC_DbgPrintf(TEXT("Warning: Cannot access directory of ")
              TEXT("\"%s\" (Please verify its DACL!)\n"), pszFileName);
            return TRUE;
        } else if (GetLastError() == ERROR_ACCESS_DENIED) {
            TC_DbgPrintf(TEXT("Warning: Cannot access ")
              TEXT("\"%s\" (Please verify its DACL!)\n"), pszFileName);
            return TRUE;
        } else {
            TC_DbgPrintf(TEXT("Cannot find folder \"%s\"!\n"), pszFileName);
        }
        return FALSE;
    }

    return (dwAttr & FILE_ATTRIBUTE_DIRECTORY) != 0;
}

BOOL WINAPI TC_PathIsMountPoint(LPCTSTR pszFileName)
{
    DWORD dwAttr;

    TC_Assert(pszFileName);

    dwAttr = GetFileAttributes(pszFileName);
    if (dwAttr == 0xffffffff) return FALSE;

    return (dwAttr & FILE_ATTRIBUTE_REPARSE_POINT) != 0;
}

UINT WINAPI TC_GetDriveType(TCHAR chDriveLetter) {

    TCHAR szRootDirectory[4];

    szRootDirectory[0] = chDriveLetter;
    szRootDirectory[1] = TEXT(':');
    szRootDirectory[2] = TEXT('\\');
    szRootDirectory[3] = TEXT('\0');

    return GetDriveType(szRootDirectory);
}

BOOL WINAPI TC_VolumeIsAccessable(TCHAR chDriveLetter)
{
    UINT uDriveType;

    uDriveType = TC_GetDriveType(chDriveLetter);
    return uDriveType == DRIVE_REMOVABLE || uDriveType == DRIVE_FIXED;
}

BOOL WINAPI TC_DriveNameIsValid(LPCTSTR pszDrive)
{
    TC_Assert(pszDrive);

    if (pszDrive == NULL || *pszDrive == TEXT('\0')) return FALSE;
    if (!_istalpha(*pszDrive)) return FALSE;
    if (pszDrive[1] == TEXT('\0')) return TRUE;
    if (pszDrive[1] != TEXT(':')) return FALSE;
    if (pszDrive[2] == TEXT('\0')) return TRUE;
    if (pszDrive[2] != TEXT('\\') && pszDrive[2] != TEXT('/')) return FALSE;

    return pszDrive[3] == TEXT('\0');
}

BOOL WINAPI TC_DriveLetterIsFree(TCHAR chDriveLetter)
{
    UINT uDriveType;

    uDriveType = TC_GetDriveType(chDriveLetter);
    return uDriveType == 1;
}

BOOL WINAPI TC_UNCFileName(LPCTSTR pszFileName)
{
    TC_Assert(pszFileName);

    if (pszFileName[0] != TEXT('\\') && pszFileName[0] != TEXT('/')) {
        return FALSE;
    }
    if (pszFileName[1] != TEXT('\\') && pszFileName[1] != TEXT('/')) {
        return FALSE;
    }
    if (pszFileName[2] == TEXT('\\') || pszFileName[2] == TEXT('/')) {
        return FALSE;
    }

    return TRUE;
}

BOOL WINAPI TC_RemoteFile(LPCTSTR pszFileName)
{
    TC_Assert(pszFileName);

    if (!TC_UNCFileName(pszFileName)) return FALSE;

    if (pszFileName[2] == TEXT('?')) return FALSE;
    if (pszFileName[2] == TEXT('.') &&
      (pszFileName[3] == TEXT('\\') || pszFileName[3] == TEXT('/'))) {
        return FALSE;
    }

    return TRUE;
}

DWORD WINAPI TC_GetDriveNo(TCHAR chDriveLetter)
{
    if (chDriveLetter >= TEXT('A') && chDriveLetter <= TEXT('Z')) {
        return chDriveLetter - TEXT('A');
    }
    if (chDriveLetter >= TEXT('a') && chDriveLetter <= TEXT('z')) {
        return chDriveLetter - TEXT('a');
    }

    return 0xffffffff;
}

DWORD WINAPI TC_GetDriveBitMask(TCHAR chDriveLetter)
{
    if (chDriveLetter >= TEXT('A') && chDriveLetter <= TEXT('Z')) {
        return 1<<(chDriveLetter - TEXT('A'));
    }
    if (chDriveLetter >= TEXT('a') && chDriveLetter <= TEXT('z')) {
        return 1<<(chDriveLetter - TEXT('a'));
    }

    return 0;
}